'use strict';

Object.defineProperty(exports, '__esModule', {
  value: true,
});
exports.applyRequiredStatus = applyRequiredStatus;

var _GraphQLError = require('../error/GraphQLError.js');

var _visitor = require('../language/visitor.js');

var _definition = require('../type/definition.js');

/**
 * Implements the "Accounting For Client Controlled Nullability Designators"
 * section of the spec. In particular, this function figures out the true return
 * type of a field by taking into account both the nullability listed in the
 * schema, and the nullability providing by an operation.
 */
function applyRequiredStatus(type, nullabilityNode) {
  const typeStack = [type];

  while (
    (0, _definition.isListType)(
      (0, _definition.getNullableType)(typeStack[typeStack.length - 1]),
    )
  ) {
    const list = (0, _definition.assertListType)(
      (0, _definition.getNullableType)(typeStack[typeStack.length - 1]),
    );
    const elementType = list.ofType;
    typeStack.push(elementType);
  }

  const applyStatusReducer = {
    RequiredDesignator: {
      leave({ element }) {
        if (element) {
          return new _definition.GraphQLNonNull(
            (0, _definition.getNullableType)(element),
          );
        } // We're working with the inner-most type

        const nextType = typeStack.pop(); // There's no way for nextType to be null if both type and nullabilityNode are valid
        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion

        return new _definition.GraphQLNonNull(
          (0, _definition.getNullableType)(nextType),
        );
      },
    },
    OptionalDesignator: {
      leave({ element }) {
        if (element) {
          return (0, _definition.getNullableType)(element);
        } // We're working with the inner-most type

        const nextType = typeStack.pop(); // There's no way for nextType to be null if both type and nullabilityNode are valid
        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion

        return (0, _definition.getNullableType)(nextType);
      },
    },
    ListNullability: {
      leave({ element }) {
        let listType = typeStack.pop(); // Skip to the inner-most list

        if (
          !(0, _definition.isListType)(
            (0, _definition.getNullableType)(listType),
          )
        ) {
          listType = typeStack.pop();
        }

        if (!listType) {
          throw new _GraphQLError.GraphQLError(
            'List nullability modifier is too deep.',
            nullabilityNode,
          );
        }

        const isRequired = (0, _definition.isNonNullType)(listType);

        if (element) {
          return isRequired
            ? new _definition.GraphQLNonNull(
                new _definition.GraphQLList(element),
              )
            : new _definition.GraphQLList(element);
        } // We're working with the inner-most list

        return listType;
      },
    },
  };

  if (nullabilityNode) {
    const modified = (0, _visitor.visit)(nullabilityNode, applyStatusReducer); // modifiers must be exactly the same depth as the field type

    if (typeStack.length > 0) {
      throw new _GraphQLError.GraphQLError(
        'List nullability modifier is too shallow.',
        nullabilityNode,
      );
    }

    return modified;
  }

  return type;
}
