'use strict';

Object.defineProperty(exports, '__esModule', {
  value: true,
});
exports.RequiredStatusOnFieldMatchesDefinitionRule =
  RequiredStatusOnFieldMatchesDefinitionRule;

var _GraphQLError = require('../../error/GraphQLError.js');

var _visitor = require('../../language/visitor.js');

var _definition = require('../../type/definition.js');

/**
 * List element nullability designators need to use a depth that is the same as or less than the
 *   type of the field it's applied to.
 *
 * Otherwise the GraphQL document is invalid.
 *
 * See https://spec.graphql.org/draft/#sec-Field-Selections
 */
function RequiredStatusOnFieldMatchesDefinitionRule(context) {
  return {
    Field(node) {
      const fieldDef = context.getFieldDef();
      const requiredNode = node.required;

      if (fieldDef && requiredNode) {
        const typeDepth = getTypeDepth(fieldDef.type);
        const designatorDepth = getDesignatorDepth(requiredNode);

        if (typeDepth > designatorDepth) {
          context.reportError(
            new _GraphQLError.GraphQLError(
              'List nullability modifier is too shallow.',
              node.required,
            ),
          );
        } else if (typeDepth < designatorDepth) {
          context.reportError(
            new _GraphQLError.GraphQLError(
              'List nullability modifier is too deep.',
              node.required,
            ),
          );
        }
      }
    },
  };

  function getTypeDepth(type) {
    let currentType = type;
    let depthCount = 0;

    while (
      (0, _definition.isListType)((0, _definition.getNullableType)(currentType))
    ) {
      const list = (0, _definition.assertListType)(
        (0, _definition.getNullableType)(currentType),
      );
      const elementType = list.ofType;
      currentType = elementType;
      depthCount += 1;
    }

    return depthCount;
  }

  function getDesignatorDepth(designator) {
    const getDepth = {
      RequiredDesignator: {
        leave({ element }) {
          return element !== null && element !== void 0 ? element : 0;
        },
      },
      OptionalDesignator: {
        leave({ element }) {
          return element !== null && element !== void 0 ? element : 0;
        },
      },
      ListNullability: {
        leave({ element }) {
          return (element !== null && element !== void 0 ? element : 0) + 1;
        },
      },
    };
    return (0, _visitor.visit)(designator, getDepth);
  }
}
