import { GraphQLError } from '../../error/GraphQLError.mjs';
import { visit } from '../../language/visitor.mjs';
import {
  assertListType,
  getNullableType,
  isListType,
} from '../../type/definition.mjs';

/**
 * List element nullability designators need to use a depth that is the same as or less than the
 *   type of the field it's applied to.
 *
 * Otherwise the GraphQL document is invalid.
 *
 * See https://spec.graphql.org/draft/#sec-Field-Selections
 */
export function RequiredStatusOnFieldMatchesDefinitionRule(context) {
  return {
    Field(node) {
      const fieldDef = context.getFieldDef();
      const requiredNode = node.required;

      if (fieldDef && requiredNode) {
        const typeDepth = getTypeDepth(fieldDef.type);
        const designatorDepth = getDesignatorDepth(requiredNode);

        if (typeDepth > designatorDepth) {
          context.reportError(
            new GraphQLError(
              'List nullability modifier is too shallow.',
              node.required,
            ),
          );
        } else if (typeDepth < designatorDepth) {
          context.reportError(
            new GraphQLError(
              'List nullability modifier is too deep.',
              node.required,
            ),
          );
        }
      }
    },
  };

  function getTypeDepth(type) {
    let currentType = type;
    let depthCount = 0;

    while (isListType(getNullableType(currentType))) {
      const list = assertListType(getNullableType(currentType));
      const elementType = list.ofType;
      currentType = elementType;
      depthCount += 1;
    }

    return depthCount;
  }

  function getDesignatorDepth(designator) {
    const getDepth = {
      RequiredDesignator: {
        leave({ element }) {
          return element !== null && element !== void 0 ? element : 0;
        },
      },
      OptionalDesignator: {
        leave({ element }) {
          return element !== null && element !== void 0 ? element : 0;
        },
      },
      ListNullability: {
        leave({ element }) {
          return (element !== null && element !== void 0 ? element : 0) + 1;
        },
      },
    };
    return visit(designator, getDepth);
  }
}
